from base_st_classes import *
import rpc
import nfs4packer
try:
    from socket import timeout
    # Remove this once we have fixed rpc.py to use timeout if available
    raise ImportError
except ImportError:
    timeout = KeyboardInterrupt
    print
    print "Socket has no timeout, use interrupt to continue hung tests"

class GssSuite(NFSSuite):
    """rpcsec_gss specific tests

    These tests don't work well, since currently rpc.py has no way
    to handle a dropped reply.  Thus a dropped reply will casue the
    tester to hang until user hits ctrl-c.
    
    """

    def testBadGssSeqnum(self):
        """GSS: using an old gss_seq_num should cause dropped reply"""
        if self.ncl.sec[0] != rpc.RPCSEC_GSS:
            self.skip("This test can only be used with RPCSEC_GSS")
        res = self.ncl.compound([self.ncl.putrootfh_op()])
        self.assert_OK(res)

        old = self.ncl.gss_seq_num
        self.ncl.gss_seq_num -= 1
        success = False
        try:
            res = self.ncl.compound([self.ncl.putrootfh_op()])
        except timeout:
            success = True
        self.ncl.gss_seq_num = old
        if not success:
            self.fail("Reply returned %i" % res)

    def testInconsistentGssSeqnum(self):
        """GSS: using inconsistent gss_seq_nums should return GARBAGE_ARGS in rpc accept_stat

        See RFC2203, end of sect 5.3.3.1
        """
        if self.ncl.sec[0] != rpc.RPCSEC_GSS or \
           self.ncl.sec[1] == rpc.rpc_gss_svc_none :
            self.skip("This test can only be used with integrity or privacy service of RPCSEC_GSS")

        # Alter self.ncl.secure_data to use wrong gss_seq_num
        self.ncl.secure_data_orig = self.ncl.secure_data
        def new_secure_data(data, init=0):
            print
            print "In new_secure_data 1"
            oldseq = self.ncl.gss_seq_num
            self.ncl.gss_seq_num += 1
            data = self.ncl.secure_data_orig(data, init)
            self.ncl.gss_seq_num = oldseq
            return data
        self.ncl.secure_data = new_secure_data

        try:
            e = None
            res = self.ncl.compound([self.ncl.putrootfh_op()])
            e = "Operation erroneously succeeded"
        except rpc.RPCGarbageArgs:
            # This is what should happen
            pass
        except rpc.RPCException, e:
            pass
        except XDRError, e:
            pass
        except timeout:
            e = "Server seems to have silently dropped request, or crashed"
        # Restore original self.ncl.mkverf
        self.ncl.secure_data = self.ncl.secure_data_orig
        if e:
            self.fail(e)

    def testBadVerfChecksum(self):
        """GSS: Bad verifier checksum should return AUTH_ERROR with stat=RPCSEC_GSS_CREDPROBLEM"""
        if self.ncl.sec[0] != rpc.RPCSEC_GSS:
            self.skip("This test can only be used with RPCSEC_GSS")

        # Alter self.ncl.mkverf to produce wrong checksum
        self.ncl.mkverf_orig = self.ncl.mkverf
        def newmkverf(init=0, buf=None):
            print
            print "In newmkverf 1"
            return self.ncl.mkverf_orig(init, buf+'x')
        self.ncl.mkverf = newmkverf

        try:
            e = None
            res = self.ncl.compound([self.ncl.putrootfh_op()])
            e = "Operation erroneously succeeded"
        except rpc.RPCAuthError, stat:
            # This should happen, with stat==rpc.RPCSEC_GSS_CREDPROBLEM
            stat = int(str(stat).split()[-1])
            if stat!=rpc.RPCSEC_GSS_CREDPROBLEM:
                e = "Returned with AuthError stat %i" % stat
        except XDRError, e:
            pass
        except timeout:
            e = "Server seems to have silently dropped request, or crashed"

        # Restore original self.ncl.mkverf
        self.ncl.mkverf = self.ncl.mkverf_orig
        if e:
            self.fail(e)

    def testBadDataChecksum(self):
        """GSS: Bad data checksum should return GARBAGE_ARGS in rpc accept_stat

        See RFC2203 sect 5.3.3.4.2
        """
        if self.ncl.sec != (rpc.RPCSEC_GSS, rpc.rpc_gss_svc_integrity) :
            self.skip("This test can only be used with integrity service of RPCSEC_GSS")

        # Alter self.ncl.secure_data to produce wrong checksum
        self.ncl.secure_data_orig = self.ncl.secure_data
        def new_secure_data(data, init=0):
            print
            print "In new_secure_data 2"
            data = self.ncl.secure_data_orig(data, init)
            if data[-4]: tail = chr(0) + data[-3:]
            else: tail = chr(1) + data[-3:]
            return data[0:-4] + tail
        self.ncl.secure_data = new_secure_data

        try:
            e = None
            res = self.ncl.compound([self.ncl.putrootfh_op()])
            e = "Operation erroneously succeeded"
        except rpc.RPCGarbageArgs:
            # This is what should happen
            pass
        except rpc.RPCException, e:
            pass
        except XDRError, e:
            pass
        except timeout:
            e = "Server seems to have silently dropped request, or crashed"
        # Restore original self.ncl.mkverf
        self.ncl.secure_data = self.ncl.secure_data_orig
        if e:
            self.fail(e)

    def testBadVersion(self):
        """GSS: bad version number should return AUTH_BADCRED"""
        if self.ncl.sec[0] != rpc.RPCSEC_GSS:
            self.skip("This test can only be used with RPCSEC_GSS")

        class BadPacker(rpc.Packer):
            def __init__(self, version):
                print
                print "In BadPacker version=%d"%version
                rpc.Packer.__init__(self)
                self.bad_gss_version = version
            def pack_cred_gss(self, gss_proc, seq_num, service, handle):
                self.pack_uint(self.bad_gss_version)
                self.pack_uint(gss_proc)
                self.pack_uint(seq_num)
                self.pack_uint(service)
                self.pack_opaque(handle)
        def make_cred_gss_new(seq_num, handle, service=rpc.rpc_gss_svc_none,
                              gss_proc=rpc.RPCSEC_GSS_DATA):
            print
            print "In make_cred_gss_new"
            p = BadPacker(version)
            p.pack_cred_gss(gss_proc, seq_num, service, handle)
            return p.get_buffer()

        orig = rpc.make_cred_gss
        rpc.make_cred_gss = make_cred_gss_new
        
        # Try with version == 0
        version = 0
        try:
            e = None
            res = self.ncl.compound([self.ncl.putrootfh_op()])
            e = "Operation erroneously succeeded"
        except rpc.RPCAuthError, stat:
            # This should happen, with stat==rpc.AUTH_BADCRED
            stat = int(str(stat).split()[-1])
            if stat!=rpc.AUTH_BADCRED:
                e = "Returned with AuthError stat %i" % stat
        except XDRError, e:
            pass
        except timeout:
            e = "Server seems to have silently dropped request, or crashed"

        if e:
            rpc.make_cred_gss = orig
            self.fail(e)

        # Try with version == 2
        version = 2
        try:
            e = None
            res = self.ncl.compound([self.ncl.putrootfh_op()])
            e = "Operation erroneously succeeded"
        except rpc.RPCAuthError, stat:
            # This should happen, with stat==rpc.AUTH_BADCRED
            stat = int(str(stat).split()[-1])
            if stat!=rpc.AUTH_BADCRED:
                e = "Returned with AuthError stat %i" % stat
        except XDRError, e:
            pass
        except timeout:
            e = "Server seems to have silently dropped request, or crashed"

        rpc.make_cred_gss = orig
        if e:
            self.fail(e)

    def testHighSeqNum(self):
        """GSS: a seq_num over MAXSEQ should return RPCSEC_GSS_CTXPROBLEM"""
        if self.ncl.sec[0] != rpc.RPCSEC_GSS:
            self.skip("This test can only be used with RPCSEC_GSS")
        old = self.ncl.gss_seq_num
        self.ncl.gss_seq_num = rpc.MAXSEQ + 1
        success = False
        try:
            e = None
            res = self.ncl.compound([self.ncl.putrootfh_op()])
            e = "Operation erroneously succeeded"
        except rpc.RPCAuthError, stat:
            # This should happen, with stat==rpc.RPCSEC_GSS_CTXPROBLEM
            stat = int(str(stat).split()[-1])
            if stat!=rpc.RPCSEC_GSS_CTXPROBLEM:
                e = "Returned with AuthError stat %i" % stat
        except XDRError, e:
            pass
        except timeout:
            e = "Server seems to have silently dropped request, or crashed"

        self.ncl.gss_seq_num = old
        if e:
            self.fail(e)

    def testBadProcedure(self):
        """GSS: bad procedure number should return AUTH_BADCRED"""
        if self.ncl.sec[0] != rpc.RPCSEC_GSS:
            self.skip("This test can only be used with RPCSEC_GSS")

        def make_cred_gss_new(seq_num, handle, service=rpc.rpc_gss_svc_none,
                              gss_proc=rpc.RPCSEC_GSS_DATA):
            print
            print "In make_cred_gss_new 3"
            p = rpc.Packer()
            p.pack_cred_gss(bad_gss_proc, seq_num, service, handle)
            return p.get_buffer()

        orig = rpc.make_cred_gss
        rpc.make_cred_gss = make_cred_gss_new

        # Try with bad_gss_proc == 4
        bad_gss_proc = 4
        try:
            e = None
            res = self.ncl.compound([self.ncl.putrootfh_op()])
            e = "Operation erroneously succeeded"
        except rpc.RPCAuthError, stat:
            # This should happen, with stat==rpc.AUTH_BADCRED
            stat = int(str(stat).split()[-1])
            if stat!=rpc.AUTH_BADCRED:
                e = "Returned with AuthError stat %i" % stat
        except XDRError, e:
            pass
        except timeout:
            e = "Server seems to have silently dropped request, or crashed"

        rpc.make_cred_gss = orig
        if e:
            self.fail(e)

    def testBadService(self):
        """GSS: bad service number should return AUTH_BADCRED"""
        if self.ncl.sec[0] != rpc.RPCSEC_GSS:
            self.skip("This test can only be used with RPCSEC_GSS")

        def make_cred_gss_new(seq_num, handle, service=rpc.rpc_gss_svc_none,
                              gss_proc=rpc.RPCSEC_GSS_DATA):
            print
            print "In make_cred_gss_new 4"
            p = rpc.Packer()
            p.pack_cred_gss(gss_proc, seq_num, badservice, handle)
            return p.get_buffer()

        orig = rpc.make_cred_gss
        rpc.make_cred_gss = make_cred_gss_new

        # Try with badservice == 0
        badservice = 0
        try:
            e = None
            res = self.ncl.compound([self.ncl.putrootfh_op()])
            e = "Operation erroneously succeeded"
        except rpc.RPCAuthError, stat:
            # This should happen, with stat==rpc.AUTH_BADCRED
            stat = int(str(stat).split()[-1])
            if stat!=rpc.AUTH_BADCRED:
                e = "Returned with AuthError stat %i" % stat
        except XDRError, e:
            pass
        except timeout:
            e = "Server seems to have silently dropped request, or crashed"

        if e:
            rpc.make_cred_gss = orig
            self.fail(e)

        # Try with badservice == 4
        badservice = 4
        try:
            e = None
            res = self.ncl.compound([self.ncl.putrootfh_op()])
            e = "Operation erroneously succeeded"
        except rpc.RPCAuthError, stat:
            # This should happen, with stat==rpc.AUTH_BADCRED
            stat = int(str(stat).split()[-1])
            if stat!=rpc.AUTH_BADCRED:
                e = "Returned with AuthError stat %i" % stat
        except XDRError, e:
            pass
        except timeout:
            e = "Server seems to have silently dropped request, or crashed"

        rpc.make_cred_gss = orig
        if e:
            self.fail(e)

