import rpc
import nfs4lib
from base_st_classes import *

class ReplaySuite(NFSSuite):
    """xid replay test.
    
    """
    def setUp(self):
        NFSSuite.setUp(self)
        self.fh = None
        self.stateid = None
        self.init_connection()

    def tearDown(self):
        self.ncl.close_file(self.fh, self.stateid, error=None)
        try: self.clean_dir(self.tmp_dir)
        except SkipException, e:
            pass

    def _replay(self,ops,error=[NFS4_OK]):

        # preserve the global xid counter so we can prevent increments
        xid = rpc.G_lastxid
        res = self.ncl.do_ops(ops)
        self.assert_status(res,error)
        
        # replay
        rpc.G_lastxid = xid
        res = self.ncl.do_ops(ops)
        self.assert_status(res,error)

        # replay
        rpc.G_lastxid = xid
        res = self.ncl.do_ops(ops)
        self.assert_status(res,error)

    #
    # Testcases covering valid equivalence classes.
    #
    def testOpen(self):
        """REPLAY: Send three OPEN calls with the same XID, SEQID, check DRC
        """
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.tmp_dir)
        ops.append(self.ncl.open(self.obj_name, OPEN4_CREATE,
                                 createattrs= {FATTR4_MODE: 0777 }))
        
        self._replay(ops)
        
    def testReplayState1(self):
        """REPLAY an erroneous OPEN of a nonexistant file"""
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.tmp_dir)
        ops.append(self.ncl.open(self.vaporfilename))
        ops.append(self.ncl.getfh_op())

        self._replay(ops,error=[NFS4ERR_NOENT])

    def testReplayState2(self):
        """REPLAY an erroneous OPEN of a dir"""
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.dirfile[:-1])
        ops.append(self.ncl.open(self.dirfile[-1]))
        ops.append(self.ncl.getfh_op())

        self._replay(ops,error=[NFS4ERR_ISDIR])

    def testReplayNonState(self):
        """REPLAY an erroneous LOOKUP"""
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.vaporfile)
        ops.append(self.ncl.getfh_op())
        self._replay(ops,error=[NFS4ERR_NOENT])

    def testLock(self):
        """REPLAY a LOCK command"""
        # Create a file and partially lock it
        self.fh, self.stateid = self.ncl.create_confirm()
        lockid = self.ncl.lock_file(self.fh, self.stateid, 20)

        # Create lock ops
        ops = [self.ncl.putfh_op(self.fh)]
        lock_owner = exist_lock_owner4(lockid, 1)
        locker = locker4(FALSE, lock_owner=lock_owner)
        lockop = self.ncl.lock_op(WRITE_LT, FALSE, 0, 10, locker)
        ops.append(lockop)

        self._replay(ops)

    def testLockDenied(self):
        """REPLAY a LOCK command that fails"""
        # Create a file and partially lock a file
        self.fh, self.stateid = self.ncl.create_confirm()
        lockid = self.ncl.lock_file(self.fh, self.stateid, 20)
        self.ncl.lock_file(self.fh, self.stateid, 0, 10, owner="other")
        
        # Create lock ops
        ops = [self.ncl.putfh_op(self.fh)]
        lock_owner = exist_lock_owner4(lockid, 1)
        locker = locker4(FALSE, lock_owner=lock_owner)
        lockop = self.ncl.lock_op(WRITE_LT, FALSE, 0, 10, locker)
        ops.append(lockop)

        self._replay(ops, error=[NFS4ERR_DENIED])

    def testUnlock(self):
        """REPLAY a LOCKU command"""
        self.fh, self.stateid = self.ncl.create_confirm()
        lockid = self.ncl.lock_file(self.fh, self.stateid)

        # Create unlock ops
        ops = [self.ncl.putfh_op(self.fh)]
        ops.append(self.ncl.locku_op(READ_LT, 1, lockid, 0, 0xffffffffffffffff))
        self._replay(ops)

        sleeptime = self.getLeaseTime() * 3 / 2
        self.sleep(sleeptime)
        self._replay(ops, error=[NFS4ERR_BAD_STATEID, NFS4_OK])

    def testClose(self):
        """REPLAY a CLOSE command"""
        self.fh, self.stateid = self.ncl.create_confirm()

        seqid = self.ncl.get_seqid()
        ops = [self.ncl.putfh_op(self.fh)]
        ops.append(self.ncl.close_op(seqid, self.stateid))
        self._replay(ops)
        
        sleeptime = self.getLeaseTime() * 3 / 2
        self.sleep(sleeptime)
        self._replay(ops, error=[NFS4ERR_BAD_STATEID, NFS4_OK])

    def testBadClose(self):
        """REPLAY a CLOSE command that fails"""
        self.fh, self.stateid = self.ncl.create_confirm()

        seqid = self.ncl.get_seqid() + 1
        ops = [self.ncl.putfh_op(self.fh)]
        ops.append(self.ncl.close_op(seqid, self.stateid))
        self._replay(ops, error=[NFS4ERR_BAD_SEQID])

    def testOpenConfirm(self):
        """REPLAY an OPEN_CONFIRM command"""

        # Open file
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.tmp_dir)
        ops.append(self.ncl.open(self.obj_name, OPEN4_CREATE, UNCHECKED4,
                                 {FATTR4_MODE: 0644}))
        ops.append(self.ncl.getfh_op())
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)
        self.fh =  res.resarray[-1].arm.arm.object
        self.stateid = res.resarray[-2].arm.arm.stateid

        # Create open_confirm ops
        ops = [self.ncl.putfh_op(self.fh)]
        confirmop = self.ncl.open_confirm_op(self.stateid, self.ncl.get_seqid())
        ops.append(confirmop)
        self._replay(ops)

    def testBadOpenConfirm(self):
        """REPLAY an OPEN_CONFIRM command that fails"""

        # Open file
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.tmp_dir)
        ops.append(self.ncl.open(self.obj_name, OPEN4_CREATE, UNCHECKED4,
                                 {FATTR4_MODE: 0644}))
        ops.append(self.ncl.getfh_op())
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)
        self.fh =  res.resarray[-1].arm.arm.object
        self.stateid = res.resarray[-2].arm.arm.stateid

        # Create open_confirm ops
        seqid = self.ncl.get_seqid() + 1
        ops = [self.ncl.putfh_op(self.fh)]
        confirmop = self.ncl.open_confirm_op(self.stateid, seqid)
        ops.append(confirmop)
        self._replay(ops, error=[NFS4ERR_BAD_SEQID])
